(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.logger = factory());
}(this, (function () { 'use strict';

  function _slicedToArray(arr, i) {
    return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
  }

  function _arrayWithHoles(arr) {
    if (Array.isArray(arr)) return arr;
  }

  function _iterableToArrayLimit(arr, i) {
    if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"] != null) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  function _unsupportedIterableToArray(o, minLen) {
    if (!o) return;
    if (typeof o === "string") return _arrayLikeToArray(o, minLen);
    var n = Object.prototype.toString.call(o).slice(8, -1);
    if (n === "Object" && o.constructor) n = o.constructor.name;
    if (n === "Map" || n === "Set") return Array.from(n);
    if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
  }

  function _arrayLikeToArray(arr, len) {
    if (len == null || len > arr.length) len = arr.length;

    for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

    return arr2;
  }

  function _nonIterableRest() {
    throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var assign = Object.assign;

  var humanize = function (ms) {
    if (ms >= 1000) {
      return "".concat(ms / 1000, "s");
    }

    return "".concat(ms, "ms");
  };

  var stringToRGB = function (string) {
    var code = string.split('').reduce(function (hash, _, i) {
      return "".concat(string.charCodeAt(i) + (hash << 5) - hash);
    }, 0);
    var hex = (code & 0x00FFFFFF).toString(16);
    return "#".concat('00000'.slice(0, 6 - hex.length) + hex);
  };

  var cloneTree = function (tree) {
    return assign({}, tree, {
      attributes: assign({}, tree.attributes),
      childNodes: tree.childNodes.map(function (vTree) {
        return cloneTree(vTree);
      })
    });
  };

  var state = {
    Internals: null
  };

  var format = function (patches) {
    var Internals = state.Internals;

    if (!Internals) {
      return {};
    }

    var decodeEntities = Internals.decodeEntities,
        PATCH_TYPE = Internals.PATCH_TYPE;
    var newPatches = {
      'Node (Insert)': [],
      'Node (Remove)': [],
      'Node (Replace)': [],
      'Node value (Set)': [],
      'Attribute (Set)': [],
      'Attribute (Remove)': []
    };
    var length = patches.length;
    var i = 0;
    var uniquePatches = 0;

    while (true) {
      var patchType = patches[i];

      if (i === length) {
        break;
      }

      uniquePatches += 1;

      switch (patchType) {
        case PATCH_TYPE.SET_ATTRIBUTE:
          {
            var vTree = patches[i + 1];
            var name = patches[i + 2];
            var value = decodeEntities(patches[i + 3]);
            newPatches['Attribute (Set)'].push({
              'Node': Internals.NodeCache.get(vTree),
              'Name': name,
              'Value': value
            });
            i += 4;
            break;
          }

        case PATCH_TYPE.REMOVE_ATTRIBUTE:
          {
            var _vTree = patches[i + 1];
            var _name = patches[i + 2];
            newPatches['Attribute (Remove)'].push({
              'Node': Internals.NodeCache.get(_vTree),
              'Name': _name
            });
            i += 3;
            break;
          }

        case PATCH_TYPE.NODE_VALUE:
          {
            var _vTree2 = patches[i + 1];
            var nodeValue = patches[i + 2];
            var oldValue = patches[i + 3];
            newPatches['Node value (Set)'].push({
              'Node': Internals.NodeCache.get(_vTree2),
              'Value': nodeValue,
              'Old value': oldValue
            });
            i += 4;
            break;
          }

        case PATCH_TYPE.INSERT_BEFORE:
          {
            var _vTree3 = patches[i + 1];
            var newTree = patches[i + 2];
            var refTree = patches[i + 3];
            newPatches['Node (Insert)'].push({
              'Container': Internals.NodeCache.get(_vTree3),
              'Node': Internals.NodeCache.get(newTree),
              'Insert Before': Internals.NodeCache.get(refTree)
            });
            i += 4;
            break;
          }

        case PATCH_TYPE.REPLACE_CHILD:
          {
            var _newTree = patches[i + 1];
            var oldTree = patches[i + 2];
            newPatches['Node (Replace)'].push({
              'New node': Internals.NodeCache.get(_newTree),
              'New tree': _newTree,
              'Old tree': oldTree
            });
            i += 3;
            break;
          }

        case PATCH_TYPE.REMOVE_CHILD:
          {
            var _vTree4 = patches[i + 1];
            newPatches['Node (Remove)'].push({
              'Tree': _vTree4
            });
            i += 2;
            break;
          }
      }
    }

    return [newPatches, uniquePatches];
  };
  /**
   * Re-usable log function. Used for during render and after render.
   *
   * @param message - Prefix for the console output.
   * @param method - Which console method to call
   * @param color - Which color styles to use
   * @param date - A date object to render
   * @param transaction - Contains: domNode, oldTree, newTree, patches, promises
   * @param options - Middleware options
   */


  var log = function (message, method, color, date, transaction, completed) {
    var domNode = transaction.domNode,
        oldTree = transaction.oldTree,
        newTree = transaction.newTree,
        patches = transaction.patches,
        promises = transaction.promises,
        options = transaction.options,
        markup = transaction.markup,
        state = transaction.state; // Shadow DOM rendering...

    if (domNode.host) {
      var ctorName = domNode.host.constructor.name;
      console[method]("%c".concat(ctorName, " render ").concat(completed ? 'ended' : 'started'), "color: ".concat(stringToRGB(ctorName)).concat(completed ? '; opacity: 0.5' : ''), completed ? completed : '');
    } else {
      console[method](message, color, completed ? completed : '');
    }

    if (!completed && domNode) {
      console.log('%cdomNode %O', 'font-weight: bold; color: #333', domNode);
    }

    if (!completed && markup) {
      console.log('%cmarkup %O', 'font-weight: bold; color: #333', markup);
    }

    if (!completed && options) {
      console.log('%coptions', 'font-weight: bold; color: #333', options);
    }

    if (state) {
      console.log('%cstate %O', 'font-weight: bold; color: #333', state);
    }

    if (transaction._cloneOldTree || newTree) {
      console.log('%coldTree %O newTree %O', 'font-weight: bold; color: #333', transaction._cloneOldTree, cloneTree(newTree));
    }

    if (patches && patches.length) {
      var _format = format(patches),
          _format2 = _slicedToArray(_format, 2),
          formattedPatches = _format2[0],
          uniqueCount = _format2[1];

      console.log('%cpatches (%d) %O', 'font-weight: bold; color: #333', uniqueCount, formattedPatches);
    } // Don't clutter the output if there aren't any promises.


    if (promises && promises.length) {
      console.log('%ctransition promises %O', 'font-weight: bold; color: #333', promises);
    }
  };

  var logger = function (_ref) {
    var _ref$minimize = _ref.minimize,
        minimize = _ref$minimize === void 0 ? false : _ref$minimize;
    return assign(function loggerTask(transaction) {
      var start = new Date();
      log('%cdiffHTML...render transaction started', 'groupCollapsed', 'color: #FF0066', start, transaction);
      var oldTree = transaction.oldTree;

      if (transaction.state.isRendering) {
        console.groupEnd();
        log('%cdiffHTML...render transaction aborted  ', minimize ? 'groupCollapsed' : 'group', 'color: #FF78B2', new Date(), transaction);
        console.groupEnd();
      }

      transaction._cloneOldTree = oldTree && cloneTree(oldTree);
      /**
       * Transaction is effectively done, but we need to listen for it to actually
       * be finished.
       */

      return function (Internals) {
        // Transaction has fully completed.
        transaction.onceEnded(function () {
          console.groupEnd();
          log('%cdiffHTML...render transaction ended  ', minimize ? 'groupCollapsed' : 'group', 'color: #FF78B2', new Date(), transaction, ' >> Completed in: ' + humanize(Date.now() - start));
          console.groupEnd();
        });
      };
    }, {
      subscribe: function (Internals) {
        state.Internals = Internals;
      },
      unsubscribe: function () {
        state.Internals = null;
      }
    });
  };

  var index = (function () {
    var opts = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    return logger(opts);
  });

  return index;

})));
